import os
import sys
import time
import math
import yaml
import copy
import torch
import wandb
import warnings
import traceback
from tqdm import tqdm
import torch.nn as nn
from torch.utils.data import DataLoader
from torch.amp import autocast, GradScaler
from contextlib import nullcontext
from dataclasses import dataclass
from .utils import *
from .models import initialize_model
from .schedulers import initialize_scheduler
from .datasets import initialize_dataset

# === Metric Logging ===

@dataclass
class Metric:
    prefix: str = None
    name: str = ""
    value: float = 0.0
    reset_value: float = 0.0
    batch_avg: bool = True
    
    def __eq__(self, metric):
        if isinstance(metric, Metric):
            return self.name == metric.name
        return False

    def __str__(self, prefixed_name=False):
        if prefixed_name:
            return f"{self.get_prefixed_name()}: {self.value:,.4g}"
        return f"{self.name}: {self.value:,.4g}"
    
    def accumulate(self, value):
        if self.batch_avg:
            self.value += value
        else:
            self.value = value
    
    def rescale(self, count):
        if self.batch_avg:
            self.value /= count
    
    def reset(self):
        self.value = self.reset_value
    
    def get_prefixed_name(self, prefix=""):
        prefix = self.prefix if self.prefix is not None else prefix
        return prefix + self.name

class MetricCollection:
    def __init__(self, metrics=None):
        if metrics is None:
            self.metrics = {}
        elif isinstance(metrics, Metric):
            self.metrics = {metrics.name: metrics}
        elif isinstance(metrics, list):
            self.metrics = {metric.name: metric for metric in metrics if isinstance(metric, Metric)}
        elif isinstance(metrics, dict):
            self.metrics = {name: metric for name, metric in metrics.items() if isinstance(metric, Metric)}
        else:
            raise TypeError(f"Expected Metric or list of Metrics, got {type(metrics)}")
    
    def __getitem__(self, name):
        return self.metrics.get(name, None)
    
    def __len__(self):
        return len(self.metrics)
    
    def __iter__(self):
        return iter(self.metrics.values())
    
    def __str__(self):
        return ", ".join(str(metric) for metric in sorted(self.metrics.values(), key=lambda m: str(m)))
    
    def __add__(self, other):
        if isinstance(other, Metric):
            return MetricCollection(metrics={**self.metrics, other.name: other})
        elif isinstance(other, MetricCollection):
            return MetricCollection(metrics={**self.metrics, **other.metrics})
        else:
            raise TypeError(f"Expected Metric or MetricCollection, got {type(other)}")
    
    def append(self, metric):
        self.add_metric(metric)
    
    def add_metric(self, metric):
        if metric is None: return
        if isinstance(metric, Metric):
            self.metrics[metric.name] = metric
        else:
            raise TypeError(f"Expected Metric or list of Metrics, got {type(metric)}")
    
    def add_metrics(self, metrics):
        if metrics is None: return
        for metric in metrics:
            self.add_metric(metric)
    
    def accumulate_metrics(self, new_metrics):
        if new_metrics is None: return
        if isinstance(new_metrics, Metric): new_metrics = [new_metrics]
        for metric in new_metrics:
            if self.metrics.get(metric.name) is not None:
                self.metrics[metric.name].accumulate(metric.value)
            else:
                self.add_metric(metric)

    def rescale_metrics(self, count):
        for metric in self.metrics.values():
            metric.rescale(count)

    def reset_metrics(self):
        for metric in self.metrics.values():
            metric.reset()
    
    def to_dict(self, prefix=""):
        return {metric.get_prefixed_name(prefix): metric.value for metric in self.metrics.values()}

def default_log_fn(output, data, target):
    return []

# === Data Augmentation ===

def default_data_fn(data, target, model, dataset):
    return data, target

# === Training ===

def train_epoch(epoch, train_loader, model, optimizer, loss_fn, log_fn=default_log_fn, data_fn=default_data_fn,
                scheduler=None, device="cpu", completed_steps=0, train_steps=None,
                checkpoint_dir="", model_name=None, val_loader=None,
                wandb_logging=False, wandb_metrics=["acc", "loss"],
                grad_clip_norm=None, accumulation_steps=1,
                mixed_precision=False, loss_backoff=InvalidLossBackoff(10, "consecutive"),
                checkpoint_freq=None, val_freq=None, info_freq=None):
    # Default model name
    if model_name is None: model_name = model.__class__.__name__
    model.train()
    accumulated_batch_metrics = MetricCollection()
    iterable = tqdm(train_loader, desc=f"Train Epoch {epoch}", leave=False, bar_format='{desc}: [{n_fmt}/{total_fmt}] {percentage:.0f}%|{bar}| [{rate_fmt}] {postfix}')
    scaler = GradScaler(device=device) if mixed_precision else None
    optimizer.zero_grad()
    for batch_idx, (data, target) in enumerate(iterable):
        with autocast(device_type=device) if mixed_precision else nullcontext():
            # Forward pass
            data = data.to(device)
            target = target.to(device)
            data, target = data_fn(data, target, model=model, dataset=train_loader.dataset)
            
            output = model(data)
            
            # Loss
            loss = loss_fn(data, output, target)
            
            # Check for invalid loss and param values
            if loss_backoff.step(loss):
                any_bad_params = report_bad_params(model)
                warnings.warn(f"Detected Invalid Loss: Epoch {epoch}, Batch {batch_idx}", RuntimeWarning)
                if any_bad_params: raise RuntimeError("Invalid values detected in model parameters.")
            
            # Accumulate metrics
            with torch.no_grad():
                new_metrics = log_fn(
                    loss=loss,
                    output=output,
                    data=data,
                    target=target,
                )
                new_metrics.append(Metric(name="loss", value=loss.item(), reset_value=0.0, batch_avg=True))
                accumulated_batch_metrics.accumulate_metrics(new_metrics)
        
        # Backward pass and gradient accumulation if applicable
        loss = loss / accumulation_steps
        loss.backward() if not mixed_precision else scaler.scale(loss).backward()
        
        if (batch_idx + 1) % accumulation_steps == 0:
            if grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=grad_clip_norm)
            optimizer.step() if not mixed_precision else scaler.step(optimizer)
            if mixed_precision: scaler.update()
            optimizer.zero_grad()
            if scheduler: scheduler.step()
            
            # WandB logging
            accumulated_batch_metrics.rescale_metrics(accumulation_steps)
            lr = scheduler.get_last_lr()[0] if scheduler is not None else optimizer.param_groups[0]["lr"]
            accumulated_batch_metrics.add_metric(
                Metric(name="lr", prefix="misc/", value=lr, batch_avg=False),
            )
            if hasattr(train_loader.dataset, "len"):
                accumulated_batch_metrics.add_metric(
                    Metric(name="seq_len", prefix="misc/", value=train_loader.dataset.len, reset_value=0.0, batch_avg=False)
                )
            if wandb_logging:
                log_data = accumulated_batch_metrics.to_dict(prefix="train/")
                wandb.log(log_data)
            
            # Post info
            if info_freq is not None and completed_steps % info_freq == 0 and completed_steps > 0:
                tqdm.write(f'Train Epoch {epoch}: [{batch_idx}/{len(train_loader)}] {accumulated_batch_metrics}')
            
            # Checkpoint
            if checkpoint_freq is not None and completed_steps % checkpoint_freq == 0 and completed_steps > 0:
                checkpoint(model_name, checkpoint_dir, model, optimizer, scheduler)
            
            # Validation
            if val_loader is not None and val_freq is not None and completed_steps % val_freq == 0 and completed_steps > 0:
                val_epoch(model, val_loader, loss_fn=loss_fn, log_fn=log_fn, data_fn=data_fn, device=device, wandb_logging=wandb_logging, wandb_metrics=wandb_metrics)
                model.train()
            
            accumulated_batch_metrics.reset_metrics()
            completed_steps += 1
        
        if train_steps is not None and completed_steps >= train_steps: break
    
    # Account for last accumulated batch
    if (batch_idx + 1) % accumulation_steps != 0:
        completed_steps += 1
        if grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=grad_clip_norm)
        optimizer.step() if not mixed_precision else scaler.step(optimizer)
        if mixed_precision: scaler.update()
        optimizer.zero_grad()
        if scheduler: scheduler.step()
        
        # WandB logging
        accumulated_batch_metrics.rescale_metrics((batch_idx % accumulation_steps) + 1)
        lr = scheduler.get_last_lr()[0] if scheduler is not None else optimizer.param_groups[0]["lr"]
        accumulated_batch_metrics.add_metric(
            Metric(name="lr", prefix="misc/", value=lr, batch_avg=False),
        )
        if hasattr(train_loader.dataset, "len"):
            accumulated_batch_metrics.add_metric(
                Metric(name="seq_len", prefix="misc/", value=train_loader.dataset.len, reset_value=0.0, batch_avg=False)
            )
        if wandb_logging:
            wandb.log(accumulated_batch_metrics.to_dict(prefix="train/"))
    
    # Step sequence length if applicable
    if hasattr(train_loader.dataset, "step"):
        train_loader.dataset.step()
    
    return completed_steps

@ torch.no_grad()
def val_epoch(model, val_loader, loss_fn, log_fn=default_log_fn, data_fn=default_data_fn,
              device="cpu",
              wandb_logging=False,):
    model.eval()
    val_metrics = MetricCollection()
    start = time.time()
    iterable = val_loader
    for data, target in iterable:
        data = data.to(device)
        target = target.to(device)
        data, target = data_fn(data, target, model=model, dataset=val_loader.dataset)
        output = model(data)
        loss = loss_fn(data, output, target)
        new_metrics = log_fn(
            loss=loss,
            output=output,
            data=data,
            target=target,
        )
        new_metrics.append(Metric(name="loss", value=loss, reset_value=0.0, batch_avg=True))
        val_metrics.accumulate_metrics(new_metrics)

    total_time = time.time() - start
    val_metrics.rescale_metrics(len(val_loader))
    
    tqdm.write(f'\033[93mVal Epoch: {val_metrics}, Elapsed: {total_time:.3f}s\033[0m')
    if wandb_logging:
        wandb.log(val_metrics.to_dict(prefix="val/"))

@ torch.no_grad()
def test_epoch(model, test_loader, loss_fn, log_fn=default_log_fn, data_fn=default_data_fn,
               device="cpu",
               wandb_logging=False,):
    model.eval()
    test_metrics = MetricCollection()
    start = time.time()
    tqdm.write("")
    iterable = tqdm(test_loader, desc=f"Test Epoch", leave=False, bar_format='\033[92m{desc}: [{n_fmt}/{total_fmt}] {percentage:.0f}%|{bar}| [{rate_fmt}] {postfix}\033[0m')
    for data, target in iterable:
        data = data.to(device)
        target = target.to(device)
        data, target = data_fn(data, target, model=model, dataset=test_loader.dataset)
        output = model(data)
        loss = loss_fn(data, output, target)
        new_metrics = log_fn(
            loss=loss,
            output=output,
            data=data,
            target=target,
        )
        new_metrics.append(Metric(name="loss", value=loss, reset_value=0.0, batch_avg=True))
        test_metrics.accumulate_metrics(new_metrics)

    total_time = time.time() - start
    test_metrics.rescale_metrics(len(test_loader))
    
    tqdm.write(f'\033[92mTest Epoch: {test_metrics}, Elapsed: {total_time:.3f}s\033[0m\n')
    if wandb_logging:
        wandb.log(test_metrics.to_dict(prefix="test/"))

def train(epochs, train_steps, benchmark_name, model, train_loader, optimizer, loss_fn, log_fn=default_log_fn, data_fn=default_data_fn,
          scheduler=None, device="cpu",
          train_config=None, mixed_precision=False,
          checkpoint_dir="", model_name=None,
          val_loader=None, test_loader=None,
          wandb_logging=True, wandb_entity=None, wandb_project=None,
          grad_clip_norm=None, accumulation_steps=1,
          loss_backoff=InvalidLossBackoff(10, "consecutive"),
          checkpoint_freq=None, val_freq=None, info_freq=None):
    try:
        sys.stdout.write("\033[?25l")
        
        NoEcho.disable_echo()
        
        # Default model name
        if model_name is None: model_name = model.__class__.__name__
        
        print(f'\033[1m{benchmark_name} Benchmark\033[0m')
        print(f'\033[1m{model_name}\033[0m')
        print(f'\033[4mTotal params: {count_parameters(model):,}\033[0m\n')
        
        # WandB Initialization
        if wandb_logging:
            assert wandb_entity is not None, "WandB entity is required for logging."
            assert wandb_project is not None, "WandB project is required for logging."
            wandb.init(
                settings=wandb.Settings(silent=True),
                mode="online" if online() else "offline",
                entity=wandb_entity,
                project=wandb_project,
                name=model_name,
                config=train_config,
            )
        
        # Use multiple GPUs if available
        if torch.cuda.device_count() > 1: model = nn.DataParallel(model)
        
        # Allocate dynamic memory if applicable for dataset
        if hasattr(train_loader.dataset, "seq_len_range"):
            min_len, max_len = train_loader.dataset.seq_len_range()
            model = allocate_dynamic_memory(model, train_loader.batch_size, min_len, max_len, device)
        else:
            # Compile the model for faster training
            model = compile_model(model, train_loader.dataset[0][0].shape, device)
        
        # Train loop
        if epochs is not None:
            for epoch in range(1, epochs + 1):
                # Train epoch
                train_epoch(
                    epoch=epoch, train_loader=train_loader, val_loader=val_loader,
                    model=model, optimizer=optimizer, scheduler=scheduler,
                    loss_fn=loss_fn, log_fn=log_fn, data_fn=data_fn, device=device,
                    checkpoint_dir=checkpoint_dir, model_name=model_name,
                    wandb_logging=wandb_logging,
                    grad_clip_norm=grad_clip_norm, accumulation_steps=accumulation_steps,
                    mixed_precision=mixed_precision, loss_backoff=loss_backoff,
                    checkpoint_freq=checkpoint_freq, val_freq=val_freq, info_freq=info_freq
                )
                
                # Test epoch
                if test_loader:
                    test_epoch(
                        model=model, test_loader=test_loader, loss_fn=loss_fn, log_fn=log_fn, data_fn=data_fn,
                        device=device,
                        wandb_logging=wandb_logging,
                    )
                
                # Model checkpoint
                checkpoint(model_name=model_name, checkpoint_dir=checkpoint_dir, model=model, optimizer=optimizer, scheduler=scheduler)
        
        elif train_steps is not None:
            completed_steps = 0
            epoch = 1
            while completed_steps < train_steps:
                completed_steps = train_epoch(
                    epoch=epoch, completed_steps=completed_steps, train_steps=train_steps, train_loader=train_loader, val_loader=val_loader,
                    model=model, optimizer=optimizer, scheduler=scheduler,
                    loss_fn=loss_fn, log_fn=log_fn, data_fn=data_fn, device=device,
                    checkpoint_dir=checkpoint_dir, model_name=model_name,
                    wandb_logging=wandb_logging,
                    grad_clip_norm=grad_clip_norm, accumulation_steps=accumulation_steps,
                    mixed_precision=mixed_precision, loss_backoff=loss_backoff,
                    checkpoint_freq=checkpoint_freq, val_freq=val_freq, info_freq=info_freq
                )
                
                # Test epoch
                if test_loader:
                    test_epoch(
                        model=model, test_loader=test_loader, loss_fn=loss_fn, log_fn=log_fn, data_fn=data_fn,
                        device=device,
                        wandb_logging=wandb_logging,
                    )
                
                # Model checkpoint
                checkpoint(model_name=model_name, checkpoint_dir=checkpoint_dir, model=model, optimizer=optimizer, scheduler=scheduler)
                
                epoch += 1
        
        # Finalize logging
        if wandb_logging: wandb.finish()
    
    except KeyboardInterrupt as e: raise e
    except Exception as e:
        if wandb_logging: wandb.finish()
        raise e
    finally:
        NoEcho.enable_echo()
        if wandb_logging: cleanup_wandb(wandb_entity, wandb_project)
        sys.stdout.write("\033[?25h")

def train_from_config_file(yaml_path, loss_fn, log_fn=default_log_fn, data_fn=default_data_fn, device="cpu"):
    """
    **Config file options:**
        `global`:
            `benchmark_name`: Name of the benchmark.
            
            `checkpoint_dir`: Directory to save model checkpoints.
            
            `dataset`:
                `name`: Class name of the dataset to use.

                `splits`: Dictionary of dataset splits with their configurations. (e.g., "train", "val", "test")
            
            `logging` (optional):
                `info_freq` (default: 100): Frequency of CLI logging training information. No CLI logging if unspecified.
                
                `wandb`: WandB logging configurations.
                    `entity`: WandB entity name.
                
                    `project`: WandB project name.
                    
            `val_freq` (default: 500): Frequency of validation during training. Ignored in no validation dataset is provided.
            
            `checkpoint_freq` (default: 500): Frequency of saving model checkpoints. No checkpointing if set to None.
        
        `experiments`:
            **List item format:**
                `general`:
                    `seed` (default: 0): Random seed for reproducibility.
                    
                    `batch_size` (default: 32): Batch size for training.
                    
                    `accumulation_steps` (default: 1): Number of batches to accumulate gradients for.
                    
                    `train_steps` (optional): Number of training steps. **(Mutually exclusive with epochs)**
                    
                    `epochs` (optional): Number of epochs to train. **(Mutually exclusive with train_steps)**
                    
                    `use_loss_fn` (default: 0):
                        If loss_fn is a list, this specifies which function to use from the list.
                        0 means use the first function in the list.
                    
                    `use_log_fn` (default: 0):
                        If log_fn is a list, this specifies which function to use from the list.
                        -1 means use default_log_fn which only logs the loss.
                    
                    `use_data_fn` (default: -1):
                        If data_fn is a list, this specifies which function to use from the list.
                        -1 means use default_data_fn which doesn't modify the data.
                    
                    `grad_clip_norm` (optional): Gradient clipping norm. No clipping if unspecified.
                    
                    `loss_backoff_count` (default: 10): Number of invalid loss backoffs before stopping training.
                    
                    `loss_backoff_type` (default: consecutive): Type of invalid loss backoff.
                    
                    `load_checkpoint` (default: False): Whether to attempt loading model from checkpoint.
                    
                    `mixed_precision` (default: False): Whether to use mixed precision for training.
                    
                    `num_workers` (default: 0): Number of workers for data loading.
                
                `model`:
                    `name`: Model class name.
                    
                    Model arguments...
                
                `optimizer`:
                    `name`: Optimizer class name (e.g., "SGD", "Adam", "AdamW").
                    
                    `exclude_weight_decay` (default ["bias", "norm"]):
                        List of parameter names to exclude from weight decay.
                        Provide empty list to not exclude any parameters.
                    
                    Optimizer arguments...
                
                `scheduler` (optional):
                    `name`: Scheduler class name (e.g., "ConstantLR", "LinearLR", "CosineAnnealingLR").
                    
                    Scheduler arguments...
    
    Args:
        yaml_path (str):
            Path to YAML configuration file.

        loss_fn (Callable):
            Args: (model_output, target)
            Returns: loss
            A function or list of functions that each compute model loss.
            If a list, `use_log_fn` in experiment general config specifies which function to use.
            If a single function, it will be treated as a list with a single item.
            At least one loss function must be provided.
            Defaults to the first function in the list if provided.
        
        log_fn (Callable, optional):
            Args: (loss, output, data, target)
            Returns: list of Metric objects or MetricCollection
            A function or list of functions that each log arbitrary metrics during training.
            If a list, `use_log_fn` in experiment general config specifies which function to use.
            If a single function, it will be treated as a list with a single item.
            Defaults to a function that logs nothing.
            **Note**: loss, lr, and sequence length (if applicable) are logged automatically.
        
        data_fn (Callable, optional):
            Args: (data, target, model, dataset)
            Returns: (data, target)
            A function or list of functions that each augment data and target before passing into model.
            If a list, `use_data_fn` in experiment general config specifies which function to use.
            If a single function, it will be treated as a list with a single item.
            Defaults to a function that does not modify the data.
        
        device (str, optional):
            Device to run training on. Defaults to cpu.
    """
    os.system('clear')
    
    if not isinstance(loss_fn, list): loss_fn = [loss_fn]
    if not isinstance(log_fn, list): log_fn = [log_fn]
    if not isinstance(data_fn, list): data_fn = [data_fn]
    
    with open(yaml_path, 'r') as f:
        configs = yaml.safe_load(f)
    
    # Extract global training configurations
    global_config = configs.get("global")
    benchmark_name = global_config.get("benchmark_name")
    checkpoint_dir = global_config.get("checkpoint_dir")
    
    # Initialize datasets
    dataset_config = global_config.get("dataset")
    dataset_name = dataset_config.get("name")
    dataset_splits = dataset_config.get("splits", {})
    if "train" in dataset_splits:
        dataset_args = dataset_splits["train"]
        train_dataset = initialize_dataset(dataset_name, **dataset_args)
    val_dataset = None
    if "val" in dataset_splits:
        dataset_args = dataset_splits["val"]
        val_dataset = initialize_dataset(dataset_name, **dataset_args)
    test_dataset = None
    if "test" in dataset_splits:
        dataset_args = dataset_splits["test"]
        test_dataset = initialize_dataset(dataset_name, **dataset_args)
    
    # Get logging configurations
    logging_config = global_config.get("logging", {})
    info_freq = logging_config.get("info_freq", 100)
    wandb_config = logging_config.get("wandb", {})
    wandb_logging = wandb_config is not None
    wandb_entity = wandb_config.get("entity")
    wandb_project = wandb_config.get("project")
    
    # Get checkpointing configurations
    val_freq = global_config.get("val_freq", 500)
    checkpoint_freq = global_config.get("checkpoint_freq", 500)
    
    # Run all experiments
    successful_count = 0
    experiments = configs.get("experiments")
    for i, experiment in enumerate(experiments):
        successful = True
        try:
            print(f'\033[1mRunning Experiment [{i + 1}/{len(experiments)}]\033[0m\n')
            
            # Reset dataset sequence lengths if applicable
            if hasattr(train_dataset, "reset"): train_dataset.reset()
            if hasattr(val_dataset, "reset"): val_dataset.reset()
            if hasattr(test_dataset, "reset"): test_dataset.reset()
            
            general_config = copy.deepcopy(experiment.get("general"))
            general_config = try_to_float(general_config)
            
            # Set seed
            seed = general_config.get("seed", 0)
            torch.manual_seed(seed)
            if device == "cuda":
                torch.cuda.manual_seed_all(seed)
            
            batch_size = general_config.get("batch_size", 32)
            accumulation_steps = general_config.get("accumulation_steps", 1)
            epochs = general_config.get("epochs", None)
            train_steps = general_config.get("train_steps", None)
            assert not (train_steps is None and epochs is None), "Either train_steps or epochs must be specified."
            assert not (train_steps is not None and epochs is not None), "Only one of train_steps or epochs can be specified."
            
            grad_clip_norm = general_config.get("grad_clip_norm", None)
            mixed_precision = general_config.get("mixed_precision", False) and device=="cuda"
            
            loss_backoff_count = general_config.get("loss_backoff_count", 10)
            loss_backoff_type = general_config.get("loss_backoff_type", "consecutive")
            loss_backoff = InvalidLossBackoff(loss_backoff_count, loss_backoff_type)
            
            # Initialize dataloaders
            num_workers = general_config.get("num_workers", 0)
            train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, num_workers=num_workers, persistent_workers=(num_workers>0), pin_memory=True, prefetch_factor=2 if num_workers > 0 else None)
            val_loader = None
            test_loader = None
            if val_dataset: val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False, num_workers=num_workers, persistent_workers=(num_workers>0), pin_memory=True, prefetch_factor=2 if num_workers > 0 else None)
            if test_dataset: test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False, num_workers=num_workers, persistent_workers=(num_workers>0), pin_memory=True, prefetch_factor=2 if num_workers > 0 else None)
            
            # Choose loss function
            loss_fn_index = general_config.get("use_loss_fn", 0)
            assert 0 <= loss_fn_index < len(loss_fn), f"Invalid loss_fn index: {loss_fn_index}. Must be between 0 and {len(loss_fn) - 1}."
            experiment_loss_fn = loss_fn[loss_fn_index]
            
            # Choose logging function
            log_fn_index = general_config.get("use_log_fn", 0)
            if log_fn_index == -1:
                experiment_log_fn = default_log_fn
            else:
                assert 0 <= log_fn_index < len(log_fn), f"Invalid log_fn index: {log_fn_index}. Must be between 0 and {len(log_fn) - 1}."
                experiment_log_fn = log_fn[log_fn_index]
            
            # Choose data function
            data_fn_index = general_config.get("use_data_fn", -1)
            if data_fn_index == -1:
                experiment_data_fn = default_data_fn
            else:
                assert 0 <= data_fn_index < len(data_fn), f"Invalid data_fn index: {data_fn_index}. Must be between 0 and {len(data_fn) - 1}."
                experiment_data_fn = data_fn[data_fn_index]
            
            model_config = copy.deepcopy(experiment.get("model"))
            model_name = model_config.pop("name")
            model_config = try_to_float(model_config)
            model_args = model_config.copy()
            model_args["device"] = device
            
            # Initialize model
            model = initialize_model(model_name, **model_args)
            
            # Initialize optimizer
            def initialize_optimizer(name, *args, **kwargs):
                optimizer_class = getattr(sys.modules["torch.optim"], name, None)
                return optimizer_class(*args, **kwargs)
            optimizer_config = copy.deepcopy(experiment.get("optimizer"))
            optimizer_name = optimizer_config.pop("name")
            optimizer_config = try_to_float(optimizer_config)
            weight_decay = float(optimizer_config.get("weight_decay", 0.0))
            exclude_weight_decay = optimizer_config.pop("exclude_weight_decay", None)
            apply_weight_decay_args = dict(
                model=model,
                weight_decay=weight_decay,
            )
            if exclude_weight_decay is not None: apply_weight_decay_args["exclude"] = exclude_weight_decay
            optimizer_config["params"] = apply_weight_decay(**apply_weight_decay_args)
            optimizer = initialize_optimizer(optimizer_name, **optimizer_config)
            
            # Initialize scheduler if specified
            scheduler = None
            scheduler_config = copy.deepcopy(experiment.get("scheduler", {}))
            if scheduler_config:
                scheduler_name = scheduler_config.pop("name")
                scheduler_config = try_to_float(scheduler_config)
                scheduler_config["optimizer"] = optimizer
                scheduler = initialize_scheduler(scheduler_name, **scheduler_config)
            
            # Load model from checkpoint if specified
            load_from_checkpoint = general_config.get("load_checkpoint", False)
            if load_from_checkpoint:
                model, optimizer, scheduler = load_checkpoint(
                    model_name=model_name, checkpoint_dir=checkpoint_dir,
                    model=model, optimizer=optimizer, scheduler=scheduler,
                    device=device
                )
            else: print(f'\033[91mStarting from scratch\033[0m')
            
            # Collect all training configurations for logging
            train_config = model_config.copy()
            train_config.update({
                "benchmark": benchmark_name,
                "model": model_name,
                "seed": seed,
                "bsz": batch_size,
                "accumulation_steps": accumulation_steps,
                "lr": optimizer_config.get("lr"),
                "weight_decay": weight_decay,
                "grad_clip_norm": grad_clip_norm,
                "permuted": dataset_splits["train"].get("permuted"),
                "min_len": train_loader.dataset.min_len if hasattr(train_loader.dataset, "min_len") else None,
                "max_len": train_loader.dataset.max_len if hasattr(train_loader.dataset, "max_len") else None,
            })
        
            # Train the model
            train(
                epochs=epochs, train_steps=train_steps, benchmark_name=benchmark_name, model_name=model_name,
                model=model, optimizer=optimizer, scheduler=scheduler,
                loss_fn=experiment_loss_fn, log_fn=experiment_log_fn, data_fn=experiment_data_fn,
                train_config=train_config, mixed_precision=mixed_precision,
                checkpoint_dir=checkpoint_dir,
                train_loader=train_loader, val_loader=val_loader, test_loader=test_loader,
                wandb_logging=wandb_logging, wandb_entity=wandb_entity, wandb_project=wandb_project,
                grad_clip_norm=grad_clip_norm, accumulation_steps=accumulation_steps,
                loss_backoff=loss_backoff,
                checkpoint_freq=checkpoint_freq, val_freq=val_freq, info_freq=info_freq,
                device=device,
            )
        except KeyboardInterrupt:
            successful = False
            terminate = input('Terminate all experiments? (y/n): ').strip().lower()
            if terminate == 'y': break
        except Exception as e:
            successful = False
            traceback_str = traceback.format_exc()
            print(f'\033[91mExperiment [{i + 1}/{len(experiments)}] failed with error:\n{traceback_str}\033[0m\n')
        if successful:
            successful_count += 1
            print(f'\033[92mExperiment [{i + 1}/{len(experiments)}] completed successfully\033[0m\n')
        
        del general_config, model_config, optimizer_config, scheduler_config, train_config
    
    print(f'\033[1m{successful_count}/{len(experiments)} experiments completed successfully\033[0m')


import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.autograd as autograd

class MLP(nn.Module):
    def __init__(self, d_in, d_hidden, d_out, activation=F.relu, bias=True, device="cpu"):
        super().__init__()
        if d_hidden is None:
            d_hidden = d_in
        self.fc1 = nn.Linear(d_in, d_hidden, bias=bias, device=device)
        self.fc2 = nn.Linear(d_hidden, d_out, bias=bias, device=device)
        self.activation = activation
        
        nn.init.kaiming_uniform_(self.fc1.weight, nonlinearity='relu')
        nn.init.xavier_uniform_(self.fc2.weight)

    def forward(self, x):
        x = self.fc1(x)
        x = self.activation(x)
        x = self.fc2(x)
        return x

class Butterfly(nn.Module):
    def __init__(self, in_size, out_size, bias=True, complex=False,
                 increasing_stride=True, init='randn', n_blocks=1,
                 device="cpu"):
        super().__init__()
        self.in_size = in_size
        self.log_n = log_n = int(math.ceil(math.log2(in_size)))
        self.n = n = 1 << log_n
        self.out_size = out_size
        self.nstacks = int(math.ceil(out_size / self.n))
        self.complex = complex
        self.increasing_stride = increasing_stride
        self.device = device
        assert n_blocks >= 1
        self.n_blocks = n_blocks
        dtype = torch.get_default_dtype() if not self.complex else {torch.float32: torch.complex64, torch.float64: torch.complex128}[torch.get_default_dtype()]
        twiddle_shape = (self.nstacks, n_blocks, log_n, n // 2, 2, 2)
        if isinstance(init, torch.Tensor):
            self.init = None
            assert init.shape == twiddle_shape
            assert init.dtype == dtype
            self.twiddle = nn.Parameter(init.clone())
        else:
            assert init in ['empty', 'randn', 'ortho', 'identity', 'fft', 'ifft']
            self.init = init
            self.twiddle = nn.Parameter(torch.empty(twiddle_shape, dtype=dtype))
        if bias:
            self.bias = nn.Parameter(torch.empty(out_size, dtype=dtype))
        else:
            self.register_parameter('bias', None)
        self.twiddle._is_structured = True  # Flag to avoid weight decay
        self._reset_parameters()
        self.to(device)

    def _reset_parameters(self):
        """Initialize bias the same way as torch.nn.Linear."""
        if self.bias is not None:
            bound = 1 / math.sqrt(self.in_size)
            nn.init.uniform_(self.bias, -bound, bound)
        twiddle = self.twiddle
        if self.init is None or self.init == 'empty':
            return
        elif self.init == 'randn':
            # complex randn already has the correct scaling of stddev=1.0
            scaling = 1.0 / math.sqrt(2)
            with torch.no_grad():
                twiddle.copy_(torch.randn(twiddle.shape, dtype=twiddle.dtype) * scaling)
        elif self.init == 'ortho':
            twiddle_core_shape = twiddle.shape[:-2]
            if not self.complex:
                theta = torch.rand(twiddle_core_shape) * math.pi * 2
                c, s = torch.cos(theta), torch.sin(theta)
                det = torch.randint(0, 2, twiddle_core_shape, dtype=c.dtype) * 2 - 1  # Rotation (+1) or reflection (-1)
                with torch.no_grad():
                    twiddle.copy_(torch.stack((torch.stack((det * c, -det * s), dim=-1),
                                               torch.stack((s, c), dim=-1)), dim=-2))
            else:
                # Sampling from the Haar measure on U(2) is a bit subtle.
                # Using the parameterization here: http://home.lu.lv/~sd20008/papers/essays/Random%20unitary%20[paper].pdf
                phi = torch.asin(torch.sqrt(torch.rand(twiddle_core_shape)))
                c, s = torch.cos(phi), torch.sin(phi)
                alpha, psi, chi = torch.rand((3, ) + twiddle_core_shape) * math.pi * 2
                A = torch.exp(1j * (alpha + psi)) * c
                B = torch.exp(1j * (alpha + chi)) * s
                C = -torch.exp(1j * (alpha - chi)) * s
                D = torch.exp(1j * (alpha - psi)) * c
                with torch.no_grad():
                    twiddle.copy_(torch.stack((torch.stack((A, B), dim=-1),
                                               torch.stack((C, D), dim=-1)), dim=-2))
        elif self.init == 'identity':
            twiddle_eye = torch.eye(2, dtype=twiddle.dtype).reshape(1, 1, 1, 1, 2, 2)
            twiddle_eye = twiddle_eye.expand(*twiddle.shape).contiguous()
            with torch.no_grad():
                twiddle.copy_(twiddle_eye)
        elif self.init in ['fft', 'ifft']:
            assert self.complex, 'fft/ifft init requires Butterfly to be complex'
            special_fn = (self._fft if self.init == 'fft' else self._ifft)
            b_fft = special_fn(self.n, normalized=True, br_first=self.increasing_stride)
            with torch.no_grad():
                twiddle[:, 0] = b_fft.twiddle
            if self.n_blocks > 1:
                twiddle_eye = torch.eye(2, dtype=twiddle.dtype).reshape(1, 1, 1, 1, 2, 2)
                twiddle_eye = twiddle_eye.expand(*twiddle[:, 1:].shape).contiguous()
                with torch.no_grad():
                    twiddle[:, 1:] = twiddle_eye

    def _complex_reshape(self, x, *shape):
        if not x.is_complex():
            return x.reshape(*shape)
        else:
            return torch.view_as_complex(torch.view_as_real(x).reshape(*shape, 2))
        
    def _fft(self, n, normalized=False, br_first=True):
        """ Construct an nn.Module based on Butterfly that exactly performs the FFT.
        Parameters:
            n: size of the FFT. Must be a power of 2.
            normalized: if True, corresponds to the unitary FFT (i.e. multiplied by 1/sqrt(n))
            br_first: which decomposition of FFT. br_first=True corresponds to decimation-in-time.
                    br_first=False corresponds to decimation-in-frequency.
        """
        log_n = int(math.ceil(math.log2(n)))
        assert n == 1 << log_n, 'n must be a power of 2'
        factors = []
        for log_size in range(1, log_n + 1):
            size = 1 << log_size
            exp = torch.exp(-2j * math.pi * torch.arange(0.0, size // 2) / size)
            o = torch.ones_like(exp)
            twiddle_factor = torch.stack((torch.stack((o, exp), dim=-1),
                                        torch.stack((o, -exp), dim=-1)), dim=-2)
            factors.append(twiddle_factor.repeat(n // size, 1, 1))
        twiddle = torch.stack(factors, dim=0).unsqueeze(0).unsqueeze(0)
        if not br_first:  # Take conjugate transpose of the BP decomposition of ifft
            twiddle = twiddle.transpose(-1, -2).flip([2])
        # Divide the whole transform by sqrt(n) by dividing each factor by n^(1/2 log_n) = sqrt(2)
        if normalized:
            twiddle /= math.sqrt(2)
        return Butterfly(n, n, bias=False, complex=True, increasing_stride=br_first, init=twiddle)

    def _ifft(self, n, normalized=False, br_first=True):
        """ Construct an nn.Module based on Butterfly that exactly performs the inverse FFT.
        Parameters:
            n: size of the iFFT. Must be a power of 2.
            normalized: if True, corresponds to unitary iFFT (i.e. multiplied by 1/sqrt(n), not 1/n)
            br_first: which decomposition of iFFT. True corresponds to decimation-in-frequency.
                    False corresponds to decimation-in-time.
        """
        log_n = int(math.ceil(math.log2(n)))
        assert n == 1 << log_n, 'n must be a power of 2'
        factors = []
        for log_size in range(1, log_n + 1):
            size = 1 << log_size
            exp = torch.exp(2j * math.pi * torch.arange(0.0, size // 2) / size)
            o = torch.ones_like(exp)
            twiddle_factor = torch.stack((torch.stack((o, exp), dim=-1),
                                        torch.stack((o, -exp), dim=-1)), dim=-2)
            factors.append(twiddle_factor.repeat(n // size, 1, 1))
        twiddle = torch.stack(factors, dim=0).unsqueeze(0).unsqueeze(0)
        if not br_first:  # Take conjugate transpose of the BP decomposition of fft
            twiddle = twiddle.transpose(-1, -2).flip([2])
        # Divide the whole transform by sqrt(n) by dividing each factor by n^(1/2 log_n) = sqrt(2)
        if normalized:
            twiddle /= math.sqrt(2)
        else:
            twiddle /= 2
        return Butterfly(n, n, bias=False, complex=True, increasing_stride=br_first, init=twiddle)
    
    def _butterfly_multiply(self, twiddle, input, increasing_stride=True, output_size=None):
        batch_size, nstacks, input_size = input.shape
        n_blocks = twiddle.shape[1]
        log_n = twiddle.shape[2]
        n = 1 << log_n
        assert twiddle.shape == (nstacks, n_blocks, log_n, n // 2, 2, 2)
        # Pad or trim input to size n
        input = F.pad(input, (0, n - input_size)) if input_size < n else input[:, :, :n]
        output_size = n if output_size is None else output_size
        assert output_size <= n
        output = input.contiguous()
        cur_increasing_stride = increasing_stride
        for block in range(n_blocks):
            for idx in range(log_n):
                log_stride = idx if cur_increasing_stride else log_n - 1 - idx
                stride = 1 << log_stride
                # shape (nstacks, n // (2 * stride), 2, 2, stride)
                t = twiddle[:, block, idx].view(
                    nstacks, n // (2 * stride), stride, 2, 2).permute(0, 1, 3, 4, 2)
                output_reshape = output.view(
                    batch_size, nstacks, n // (2 * stride), 1, 2, stride)
                output = (t * output_reshape).sum(dim=4)
            cur_increasing_stride = not cur_increasing_stride
        return output.view(batch_size, nstacks, n)[:, :, :output_size]
    
    def forward(self, input, transpose=False, conjugate=False, subtwiddle=False):
        """
        Parameters:
            input: (batch, *, in_size)
            transpose: whether the butterfly matrix should be transposed.
            conjugate: whether the butterfly matrix should be conjugated.
            subtwiddle: allow using only part of the parameters for smaller input.
                Could be useful for weight sharing.
                out_size is set to self.nstacks * self.n in this case
        Return:
            output: (batch, *, out_size)
        """
        twiddle = self.twiddle
        output = self.pre_process(input)
        output_size = self.out_size if self.nstacks == 1 else None
        if subtwiddle:
            log_n = int(math.ceil(math.log2(input.size(-1))))
            n = 1 << log_n
            twiddle = (twiddle[:, :, :log_n, :n // 2] if self.increasing_stride
                       else twiddle[:, :, -log_n:, :n // 2])
            output_size = None
        if conjugate and self.complex:
            twiddle = twiddle.conj()
        if not transpose:
            output = self._butterfly_multiply(twiddle, output, self.increasing_stride, output_size)
        else:
            twiddle = twiddle.transpose(-1, -2).flip([1, 2])
            last_increasing_stride = self.increasing_stride != ((self.n_blocks - 1) % 2 == 1)
            output = self._butterfly_multiply(twiddle, output, not last_increasing_stride, output_size)
        if not subtwiddle:
            return self.post_process(input, output)
        else:
            return self.post_process(input, output, out_size=output.size(-1))
    
    def pre_process(self, input):
        # Reshape to (N, in_size)
        input_size = input.size(-1)
        output = self._complex_reshape(input, -1, input_size)
        batch = output.shape[0]
        output = output.unsqueeze(1).expand(batch, self.nstacks, input_size)
        return output

    def post_process(self, input, output, out_size=None):
        if out_size is None:
            out_size = self.out_size
        batch = output.shape[0]
        output = output.view(batch, self.nstacks * output.size(-1))
        if out_size != output.shape[-1]:  # Take top rows
            output = output[:, :out_size]
        if self.bias is not None:
            output = output + self.bias[:out_size]
        return output.view(*input.size()[:-1], out_size)

class GatedRMSNorm(nn.Module):
    def __init__(self, d, eps=1e-5, device=None):
        """Gated Root Mean Square Layer Normalization

        Paper: https://arxiv.org/abs/1910.07467
        """
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(d, device=device))

    def forward(self, x, z=None):
        if z is not None:
            x = x * F.silu(z)
        return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps) * self.weight

class DilatedSlidingWindow(autograd.Function):
    @staticmethod
    def forward(ctx, x, size, stride, dilation, dim, pad):
        if torch.is_grad_enabled() and x.requires_grad:
            ctx.size = size
            ctx.stride = stride
            ctx.dilation = dilation
            ctx.dim = dim
            ctx.pad = pad
            ctx.x_shape = x.shape
            ctx.save_for_backward(x)
        
        ndim = x.dim()
        if dim < -ndim or dim >= ndim:
            raise IndexError(f"Dimension out of range (expected to be in range of [-{ndim}, {ndim-1}], but got {dim})")
        
        if dim < 0:
            dim = ndim + dim

        if pad[0] > 0 or pad[1] > 0:
            pad_tuple = [0] * (2 * ndim)
            pad_idx = 2 * (ndim - 1 - dim)
            pad_tuple[pad_idx] = pad[0]
            pad_tuple[pad_idx + 1] = pad[1]
            x = F.pad(x, tuple(pad_tuple))

        n_padded = x.shape[dim]
        effective_window_size = (size - 1) * dilation + 1
        num_windows = (n_padded - effective_window_size) // stride + 1

        if num_windows <= 0:
            final_shape = list(x.shape)
            final_shape[dim:dim+1] = [0, size]
            return torch.empty(final_shape, dtype=x.dtype, device=x.device)

        out_shape = list(x.shape)
        out_shape[dim:dim+1] = [num_windows, size]

        original_strides = x.stride()
        element_stride = original_strides[dim]

        out_stride = (
            original_strides[:dim] 
            + (stride * element_stride, dilation * element_stride) 
            + original_strides[dim+1:]
        )
        y = x.as_strided(out_shape, tuple(out_stride))

        return y

    @staticmethod
    @autograd.function.once_differentiable
    def backward(ctx, grad_output):
        if not ctx.saved_tensors:
            return None, None, None, None, None, None
        
        x, = ctx.saved_tensors  # This is now the original input
        
        # Create grad_input for the original input shape
        grad_input = torch.zeros_like(x)
        
        # Apply padding to grad_input if needed (to match forward pass)
        ndim = x.dim()
        dim = ctx.dim
        if dim < 0:
            dim = ndim + dim
            
        if ctx.pad[0] > 0 or ctx.pad[1] > 0:
            pad_tuple = [0] * (2 * ndim)
            pad_idx = 2 * (ndim - 1 - dim)
            pad_tuple[pad_idx] = ctx.pad[0]
            pad_tuple[pad_idx + 1] = ctx.pad[1]
            grad_input_padded = F.pad(grad_input, tuple(pad_tuple))
        else:
            grad_input_padded = grad_input

        n_padded = grad_input_padded.shape[dim]
        effective_window_size = (ctx.size - 1) * ctx.dilation + 1
        num_windows = (n_padded - effective_window_size) // ctx.stride + 1

        if num_windows > 0:
            out_shape = list(grad_input_padded.shape)
            out_shape[dim:dim+1] = [num_windows, ctx.size]
            original_strides = grad_input_padded.stride()
            element_stride = original_strides[dim]
            out_stride = (
                original_strides[:dim] 
                + (ctx.stride * element_stride, ctx.dilation * element_stride) 
                + original_strides[dim+1:]
            )
            grad_input_padded.as_strided(out_shape, out_stride).add_(grad_output)

        # If we padded, we need to "unpad" to get back to original input size
        if ctx.pad[0] > 0 or ctx.pad[1] > 0:
            # Extract the original region from the padded gradient
            slices = [slice(None)] * ndim
            slices[dim] = slice(ctx.pad[0], grad_input_padded.shape[dim] - ctx.pad[1] if ctx.pad[1] > 0 else None)
            grad_input = grad_input_padded[tuple(slices)]
        else:
            grad_input = grad_input_padded
        
        return grad_input, None, None, None, None, None

def dilated_sliding_window(x, size, stride=1, dilation=1, dim=-1, pad=(0, 0)):
    return DilatedSlidingWindow.apply(x, size, stride, dilation, dim, pad)
